// NpMsgrMain.cpp: implementation of the CNpMsgrMain class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "NpMsgrMain.h"

#include "GBStream.h"
#include "NpSessionBase.h"

#include "PWDCrypt.h"
#include "Resource.h"

#if defined(WIN32)
// USES_CONVERSION   ߰
#include <comdef.h>
#include <CRTDBG.H>
#include <atlconv.h> 

#include <winsock2.h>
#endif // WIN32

// #ifdef _DEBUG
#include <fstream>
// #endif

#include "GTCommonAPI.h"
#include "GTCommonComponentDefine.h"

#define WM_TRACE		wm_trace

#define PORT_COUNT_FROM_10500			10

#define NP_MSNG_P2P_CMMD_PROFILE_EXCHANGE			0x7001

#define IDT_RELOGIN_START				100
#define IDT_CHECK_NETWORK				102

#if defined(WIN32)
#define MAX_WAIT_RECV_PONG_TIME			20000
#else
#define MAX_WAIT_RECV_PONG_TIME			10000
#endif

#define NETWORK_CHECK_TIME				30000


enum GTLoginServerErrorDetailCode
{
	WASCODE_OK = 100,
	WASCODE_ERROR = 200,
	WASCODE_AUTH_FAIL = 202,
	WASCODE_AUTH_TICKET_EXPIRED = 220,
	WASCODE_AUTH_INVALID_CLIENT_VERSION = 221,		// NpMsgrMain.cpp define  ƾ 
	WASCODE_AUTH_EMAIL_DUPLICATION = 222,
	WASCODE_INVALID_TENANT = 223,
	WASCODE_MTS_SERVER_ERROR = 224
};


int	IsConnected (SOCKET s, fd_set* rd, fd_set* wr, fd_set* ex)
{
	WSASetLastError( 0 );
	if ( !FD_ISSET(s, rd) && !FD_ISSET(s, wr))
		return 0;
	if ( FD_ISSET(s, ex))
		return 0;
	return 1;
}

bool CheckConnection(LPCTSTR pszcAddr,int nPort, int nTimeOut)
{
	int port  = nPort;
	bool bConnect = false;

	WSADATA wsaData;
	int err = WSAStartup( MAKEWORD( 2, 2 ), &wsaData );

	hostent * p = gethostbyname(T2CA(pszcAddr));
	if(p == NULL) return FALSE;

	char szIP[16];
	sprintf(szIP, "%d.%d.%d.%d", (BYTE)p->h_addr_list[0][0], (BYTE)p->h_addr_list[0][1], (BYTE)p->h_addr_list[0][2], (BYTE)p->h_addr_list[0][3]);

	SOCKET s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);

	if (nTimeOut > 0)
	{
		unsigned long fl = 1;
		if (ioctlsocket(s, FIONBIO, &fl) < 0) // non-blocking setting
		{
			shutdown(s, 1);
			closesocket(s);
			WSACleanup();
			return bConnect;
		}
	}

	SOCKADDR_IN          server_addr;
	server_addr.sin_family = AF_INET;
	server_addr.sin_port = htons(port); 
	server_addr.sin_addr.s_addr =inet_addr(szIP);

	int nResult = connect(s, (SOCKADDR *) &server_addr, sizeof(server_addr)); 
	if (nResult != 0)
	{
		fd_set rdevents;
		fd_set wrevents;
		fd_set exevents;
		struct timeval tv;

		FD_ZERO(&rdevents);
		FD_SET(s, &rdevents);
		wrevents = rdevents;
		exevents = rdevents;
		tv.tv_sec = nTimeOut;
		tv.tv_usec = 0;
		nResult = select (s+1, &rdevents, &wrevents, &exevents, &tv);

		if (IsConnected(s, &rdevents, &wrevents, &exevents))
		{
			bConnect = true;
		}
	}
	shutdown(s, 1);
	closesocket(s);
	WSACleanup();

	return bConnect;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CNpMsgrMain::CNpMsgrMain(_INpMsgrEvents * pEvents) : m_pEvents(pEvents)
, m_nLRSCookie(NULL)
, m_bSendPing(true)
, m_nSendPingInterval(MAX_WAIT_RECV_PONG_TIME)
, m_nReloginTryCount(0)
, m_nLastRecvPongMSec(0)
, m_bRecvPongFlag(true)
, m_bMobileFlag(false)
, m_uLangLocaleCode(82)
, m_bPrevNetworkCheckIsFalse(false)
{
	srand( (unsigned)time( NULL ) );

	wm_trace_init();

	m_ikWnd.SetWndHandler(this);
	m_ikWnd.Create( "NpMsgrMain Helper Window for gametalk", NULL, 0 );
	ASSERT( m_ikWnd.m_hWnd != NULL );

	m_mutexSession.SetRecursive();	// OnConnect() called on another thread..
	
	m_pLoginNpSession = NULL;
	m_pMainServerNpSession = NULL;

	m_pNpSession_pool = NULL;
	m_pNpMsgrUser = new CNpMsgrUser;
	m_pNpMsgrConfig = new CNpMsgrConfig;
	m_pNpMsgrBuddy = new CNpMsgrBuddy;
	m_pNpMsgrIPT = new CNpMsgrIPT;

    m_bReloginProcess = false;
    
	m_pNpSD_MainServer = new CNpSessionDispatcher( m_pNpMsgrUser );
	
	CONFIG()->InitModule( USER() );
	BUDDY()->InitModule( USER(), CONFIG(), MAINSERVER_SD() );
	IPT()->InitModule( MAINSERVER_SD() );
	
	GetMainServer_SessionDispatcher()->Register_INCOMING_REQUEST_PACKET_Handler(MSNG_CMD_SYSTEMNOTIFY, this);
	GetMainServer_SessionDispatcher()->Register_INCOMING_REQUEST_PACKET_Handler(MSNG_CMD_EXTENDNOTIFY, this);
	GetMainServer_SessionDispatcher()->Register_INCOMING_REQUEST_PACKET_Handler(MSNG_CMD_NOTIFYNOTE, this);
	GetMainServer_SessionDispatcher()->Register_INCOMING_REQUEST_PACKET_Handler(MSNG_CMD_NOTIFYUNREADNOTECNT, this);
	GetMainServer_SessionDispatcher()->Register_INCOMING_REQUEST_PACKET_Handler(MSNG_CMD_NOTIFYMEMO, this);
	GetMainServer_SessionDispatcher()->Register_INCOMING_REQUEST_PACKET_Handler(MSNG_CMD_NOTIFYREFRESH, this);

	m_nReloginTryCount = 0;
	m_strLoginCode = "";
	m_strTicketType = "";
	m_nMobileDataOverwriteMode = 0;
	
	m_bPrevLoginOKFlag = false;
}

CNpMsgrMain::~CNpMsgrMain()
{
	GetMainServer_SessionDispatcher()->Deregister_INCOMING_REQUEST_PACKET_Handler(MSNG_CMD_SYSTEMNOTIFY);
	GetMainServer_SessionDispatcher()->Deregister_INCOMING_REQUEST_PACKET_Handler(MSNG_CMD_EXTENDNOTIFY);
	GetMainServer_SessionDispatcher()->Deregister_INCOMING_REQUEST_PACKET_Handler(MSNG_CMD_NOTIFYNOTE);
	GetMainServer_SessionDispatcher()->Deregister_INCOMING_REQUEST_PACKET_Handler(MSNG_CMD_NOTIFYUNREADNOTECNT);
	GetMainServer_SessionDispatcher()->Deregister_INCOMING_REQUEST_PACKET_Handler(MSNG_CMD_NOTIFYMEMO);
	GetMainServer_SessionDispatcher()->Deregister_INCOMING_REQUEST_PACKET_Handler(MSNG_CMD_NOTIFYREFRESH);

	if( USER()->LoginStatus() != LS_LOGIN_READY )
		Logout();
	
	delete m_pNpMsgrUser;
	delete m_pNpMsgrConfig;
	delete m_pNpMsgrBuddy;
	delete m_pNpMsgrIPT;
	delete m_pNpSD_MainServer;

	//	delete m_pQueryLoginAuth;
	//	delete m_pQueryLoginGetData;
	if( m_pLoginNpSession )
	{
		m_pLoginNpSession->Delete();
		m_pLoginNpSession = NULL;
	}

	if( m_pMainServerNpSession )
	{
		m_pMainServerNpSession->Delete();
		m_pMainServerNpSession = NULL;
	}

	if( m_pNpSession_pool )
	{
		m_pNpSession_pool->Delete();
		m_pNpSession_pool = NULL;
	}
	
	m_ikWnd.DestroyWindow();
}

/////////////////////////////////////////////////////////////////////////////////
//

void CNpMsgrMain::SetMobileInfo(short nOverwriteDataType, unsigned short nCountryCode, const char * strMobileNumber, const char * strMobileToken)
{
	m_bMobileFlag = true;
	m_uLangLocaleCode = nCountryCode;
	m_nMobileDataOverwriteMode = nOverwriteDataType;
	
	if(strMobileToken)
		m_strMobileToken = strMobileToken;
	
	if(strMobileNumber)
		m_strMobileNumber = strMobileNumber;	
}


bool CNpMsgrMain::Login(LPCWSTR wszUserId, const char * szUserPassword, const char * strLoginCode, const char * strTicketType, unsigned short usOSType)
{
	nfc::AutoLock<nfc::CMutex> autolock(&m_mutexSession);
	
	m_nLRSCookie = NULL;
	m_nReloginTryCount = 0;
	m_nDnsFailReport = 0;
	
#ifdef _DEBUG
//	string strTmpUserId;
//	stl_ik_wcs2mbs( wszUserId, strTmpUserId );
//	F_TRACE("CNpMsgrMain::Login(id=%s, pwd=%s)", strTmpUserId.c_str(), szUserPassword);	
#endif

	ASSERT( USER()->m_eLoginStatus == LS_LOGIN_READY );
	ASSERT( USER()->m_eOnlineStatus == ST_OFFLINE );

	if( USER()->m_eLoginStatus != LS_LOGIN_READY )
	{
//		NTALK_LOG->Debug(L"[LOGIN] FAIL Login : LoginStatus is not LS_LOGIN_READY, [%d]", USER()->m_eLoginStatus);
		return false;
	}

	if( USER()->m_eOnlineStatus != ST_OFFLINE )
	{
//		NTALK_LOG->Debug(L"[LOGIN] FAIL Login : LoginStatus is not ST_OFFLINE, [%d]", USER()->m_eOnlineStatus);
		return false;
	}

	m_vecLoginResponseData.clear();
	m_strLastErrorMsg = "";

	USER()->m_wstrInputLoginId = wszUserId;
	USER()->m_wstrId = wszUserId;
	USER()->m_strPassword = szUserPassword;

	USER()->m_eLoginStatus = LS_LOGIN_CONNECTING;

	m_strLoginCode = strLoginCode;
	m_strTicketType = strTicketType;
	
	m_usOSType = usOSType;

	bool bResult = Internal_LoginConnect();

	if( bResult != true )
	{
		USER()->m_eLoginStatus = LS_LOGIN_READY;
		USER()->m_eOnlineStatus = ST_OFFLINE;
	}

	return bResult;
}

void CNpMsgrMain::SetLoginServerAddress(LPCWSTR strAddress, UINT nPort)
{
	m_wstrLoginServerAddr = strAddress;
	m_nLoginServerPort = nPort;
}

bool CNpMsgrMain::Internal_LoginConnect()
{
	USES_CONVERSION;

	ASSERT( USER()->m_eLoginStatus == LS_LOGIN_CONNECTING || USER()->m_eLoginStatus == LS_RELOGIN_CONNECTING );

	// connect
	if( m_pLoginNpSession )
		m_pLoginNpSession->Delete();

	m_pLoginNpSession = CreateNpSession( this, true );
    
	bool mainThreadMode = true;
    
    m_pLoginNpSession->GetIckarSocket()->SetMainThreadEventMode(mainThreadMode);

	//GT_FLOG(L"ConnectSession( %s, %d)", m_wstrLoginServerAddr.c_str(), m_nLoginServerPort);

	return m_pLoginNpSession->ConnectSession( W2A(m_wstrLoginServerAddr.c_str()), m_nLoginServerPort);
}

void CALLBACK CNpMsgrMain::OnInternalLoginResult(USHORT usResultCode, USHORT usErrorCode)
{
//	F_TRACE("CNpMsgrMain::OnInternalLoginResult(%d, %d)", usResultCode, usErrorCode);
	//GT_FLOG(L"CNpMsgrMain::OnInternalLoginResult(%d, %d)", usResultCode, usErrorCode);

	if( usResultCode == RC_OK )
	{
		USER()->m_eLoginStatus = LS_LOGIN_COMPLETE;

		if( m_pMainServerNpSession )
			m_pMainServerNpSession->Delete();
		m_pMainServerNpSession = m_pLoginNpSession;
		m_pLoginNpSession = NULL;

		m_pEvents->NpMsgr_OnLoginEnd( this, usResultCode, usErrorCode );	
		
        StartCheckNetworkProcess();

        StopReloginProcess();
		
		m_bPrevLoginOKFlag = true;
	}
	else
	{
		//NpSession_AsyncDelete( m_pLoginNpSession );
		m_pLoginNpSession->Delete();
		m_pLoginNpSession = NULL;

		if( usResultCode == RC_CONNECTFAIL && usErrorCode == 11001 )	// (11001)WSAHOST_NOT_FOUND	-- dns 
		{
			USES_CONVERSION;

			UINT nRandPort = rand() % PORT_COUNT_FROM_10500;

			m_nDnsFailReport = 1;
			m_pLoginNpSession = CreateNpSession( this, true );
			m_pLoginNpSession->ConnectSession( W2A(m_wstrLoginServerAddr.c_str()), m_nLoginServerPort + nRandPort );
			return;
		}

		USER()->m_eLoginStatus = LS_LOGIN_READY;
		USER()->m_eOnlineStatus = ST_OFFLINE;

		m_pEvents->NpMsgr_OnLoginEnd( this, usResultCode, usErrorCode );

		USER()->Clear();
		BUDDY()->ClearAll();

		MAINSERVER_SD()->EndMainServerSession();

		if( m_pMainServerNpSession )
		{
			m_pMainServerNpSession->CloseSession();

			m_pMainServerNpSession->Delete();
			m_pMainServerNpSession = NULL;
		}
		
		m_bPrevLoginOKFlag = false;
	}

	m_vecLoginResponseData.clear();
}

void CALLBACK CNpMsgrMain::OnInternalReloginResult(USHORT usResultCode, USHORT usErrorCode)
{
//	F_TRACE("CNpMsgrMain::OnInternalReloginResult(%d, %d)\r\n", usResultCode, usErrorCode);
	//GT_FLOG(L"CNpMsgrMain::OnInternalReloginResult(%d, %d)", usResultCode, usErrorCode);

	if( usResultCode != RC_OK )
	{
		m_pLoginNpSession->CloseSession();

		if( usErrorCode == WASCODE_AUTH_INVALID_CLIENT_VERSION || usErrorCode == WASCODE_AUTH_FAIL 
			|| usErrorCode == WASCODE_AUTH_TICKET_EXPIRED || usErrorCode == WASCODE_AUTH_EMAIL_DUPLICATION )
		{
			StopReloginProcess();
			return;
		}

		m_nReloginTryCount++;	// (0 = ready, 1 = first try, ++ )

		if(m_bReloginProcess)
		{
			UINT nReloginInterval = 1000 * 60;
			switch( m_nReloginTryCount )
			{
			case 10:	// 10      α׾ƿ ó(  籸   8 ҿ)
				m_nReloginTryCount = 0;

				StopReloginProcess();

				USER()->m_eLoginStatus = LS_LOGIN_READY;
				USER()->m_eOnlineStatus = ST_OFFLINE;

				m_pEvents->NpMsgr_OnDisconnected(this, 2);
				return;
			default:
				ASSERT(FALSE);
				nReloginInterval = (1000 * 60) + (::rand() % (60 * 1000));	// ּ 1 + x
				break;
			}

			OutputDebugString(_T("\r\n##### RELOGIN TIMER STARTED CNpMsgrMain::OnInternalReloginResult() #####"));
//			NTALK_LOG->Debug(L"[LOGIN] IDT_RELOGIN_START [%d,%d]", m_nReloginTryCount, nReloginInterval);

			TIMER_START( IDT_RELOGIN_START, nReloginInterval, NULL );
			USER()->m_eLoginStatus = LS_RELOGIN_WAIT_START;
		}
	}
	else
	{
//		IK_TRACE("RELOGIN SUCCESS!!!\n");

		ASSERT( m_pMainServerNpSession != NULL );
		ASSERT( m_pLoginNpSession == NULL );

		USER()->m_eLoginStatus = LS_LOGIN_COMPLETE;
		
		m_pEvents->NpMsgr_OnReloginEnd( this, usResultCode, usErrorCode );
		
		m_nReloginTryCount = 0;
		
		StartCheckNetworkProcess();
	}
}

void CNpMsgrMain::Disconnect()
{
	OutputDebugString(_T("\r\n##### CNpMsgrMain::Disconnect #####"));
	nfc::AutoLock<nfc::CMutex> autolock(&m_mutexSession);

	StopCheckNetworkProcess();
	
    StopReloginProcess();
    
	m_nReloginTryCount = 0;
	
	if( m_pLoginNpSession )
	{
		m_pLoginNpSession->CloseSession();
		
		m_pLoginNpSession->Delete();
		m_pLoginNpSession = NULL;
	}
	
	if( m_pMainServerNpSession )
	{
		m_pMainServerNpSession->CloseSession();
		
		m_pMainServerNpSession->Delete();
		m_pMainServerNpSession = NULL;
	}
	
	if(m_bPrevLoginOKFlag)
		MAINSERVER_SD()->SetClosedOnlineMode();
	else 
	{
		USER()->Clear();
		BUDDY()->ClearAll();
				
		MAINSERVER_SD()->EndMainServerSession();
	}
}


bool CNpMsgrMain::Relogin()
{
	nfc::AutoLock<nfc::CMutex> autolock(&m_mutexSession);

	//Disconnect();

	StartReloginProcess(true);
	
	return true;
}


bool CNpMsgrMain::Logout()
{
	WM_TRACE(_T("FT Enter<CNpMsgrMain::Logout"));

	m_nLRSCookie = NULL;
	m_nReloginTryCount = 0;
	m_nDnsFailReport = 0;

	if( m_pLoginNpSession )
	{
		m_pLoginNpSession->CloseSession();

		WM_TRACE(_T("FT Probe<CNpMsgrMain::Logout - delete m_pLoginNpSession"));
		m_pLoginNpSession->Delete();
		m_pLoginNpSession = NULL;
	}

	if( m_pMainServerNpSession )
	{
		MAINSERVER_SD()->Send_REQUEST_PACKET( MSNG_CMD_LOGOUT, 0, "" );
		m_pMainServerNpSession->CloseSession();

		m_pMainServerNpSession->Delete();
		m_pMainServerNpSession = NULL;
	}

	MAINSERVER_SD()->EndMainServerSession();

	if( USER()->m_eLoginStatus != LS_LOGIN_READY )
	{
		BUDDY()->Save_All_BuddyCacheData();	
	}

	m_strCustomData.clear();

	BUDDY()->ClearAll();
	USER()->Clear(); 

	//USER()->m_eOnlineStatus = ST_OFFLINE; 
	//USER()->m_eLoginStatus = LS_LOGIN_READY;

	StopCheckNetworkProcess();

	WM_TRACE(_T("FT Leave<CNpMsgrMain::Logout"));
	return false;
}

const char * CNpMsgrMain::GetLastErrorMsg()
{
	return m_strLastErrorMsg.c_str();
}


void CNpMsgrMain::SendSessionControlPacket(void)
{
	if(false == m_bRecvPongFlag)
	{
		m_pEvents->NpMsgr_OnDisconnected( this, 0);
	}
	else
	{
		m_bRecvPongFlag = false;
	}
	
	GetMainServer_SessionDispatcher()->Send_PING();
}

bool CNpMsgrMain::SendPing(void)
{
	return GetMainServer_SessionDispatcher()->PING();
}

void CNpMsgrMain::SetPingSettting(bool bSendPing, int nInterval)
{
	bool bStartFlag = (bSendPing && !m_bSendPing);
	bool bStopFlag = (!bSendPing && m_bSendPing);
	
	m_bSendPing = bSendPing;
	m_nSendPingInterval = nInterval;

	if(m_nSendPingInterval < 0 && bStartFlag)
	{
		m_nSendPingInterval = MAX_WAIT_RECV_PONG_TIME;
	}
}

void CNpMsgrMain::RecoveryNetwork()
{
//	NTALK_LOG->Debug(L"[LOGIN] IDT_CHECK_NETWORK [DISCONNECTED]");
	OutputDebugString(_T("\r\n====================== IDT_CHECK_NETWORK [DISCONNECTED] ======================\r\n"));

	m_pEvents->NpMsgr_OnDisconnected( this, 1 );
	
	if (m_pMainServerNpSession)
		m_pMainServerNpSession->CloseSession();

	OnInternalMainSessionClosed(0);
}


/**********************************************************************
Np Messenger Client  Ⱓ  Ÿ ְ   
**********************************************************************/

const char * CNpMsgrMain::GetResultCodeMsg(int nResultCode)
{
	const char *pstr = "";
	switch( nResultCode )
	{
	case RC_OK:
		pstr = "OK";
		break;

	case RC_CONNECTFAIL:
		pstr = "connect fail";
		break;
	case RC_SOCKET_CLOSED:
		pstr = "socket closed";
		break;
	case RC_SOCKET_RECV_ERROR:
		pstr = "socket recv error";
		break;
	case RC_SOCKET_SEND_ERROR:
		pstr = "socket send error";
		break;

	case RC_RECV_ERROR_RESULT:
		pstr = "receive error code";
		break;

	case RC_INVALID_DATA:
		pstr = "invalid data";
		break;

	case RC_INVALID_PROTOCOL:
		pstr = "invalid protocol";
		break;
	case RC_NOT_COMPATABLE_PROTOCOL:
		pstr = "not compatable protocol verseion";
		break;
	case RC_AUTH_FAIL:
		pstr = "auth fail";
		break;
	case RC_CANNOT_FIND_TARGET:
		pstr = "cannot find target";
		break;
	case RC_RECV_UNEXPECTED_CMMD:
		pstr = "receive unexpected cmmd";
		break;
	case RC_CANNOT_FOUND_NODE:
		pstr = "cannot found node";
		break;
	case RC_XML_PARSE_ERROR:
		pstr = "xml parse error";
		break;

	case RC_UNDEFINED:
		pstr = "undefined";
		break;

	}

	return pstr;
}

/**********************************************************************
server ʿ ִ  ڵ
**********************************************************************/
const char * GetNpIMErrorMsg(int nNpErrorCode)
{
#define CASE_RETURN_STRTYPE( val )	\
	case val: \
				  return #val;

		switch(nNpErrorCode)
		{					
			CASE_RETURN_STRTYPE( MSNG_OK )
				CASE_RETURN_STRTYPE( MSNG_OK_UPDATECACHE )
				CASE_RETURN_STRTYPE( MSNG_OK_FIRSTLOGIN )
				CASE_RETURN_STRTYPE( MSNG_OK_SAVE_MESSAGE )
				CASE_RETURN_STRTYPE( MSNG_OK_ALREADYBUDDY )

				CASE_RETURN_STRTYPE( MSNG_ERR_INVALIDDATA )
				CASE_RETURN_STRTYPE( MSNG_ERR_EXPIREDCOOKIE )
				CASE_RETURN_STRTYPE( MSNG_ERR_FAILAUTH )
				CASE_RETURN_STRTYPE( MSNG_ERR_INVALIDREQ )
				CASE_RETURN_STRTYPE( MSNG_ERR_UNREGISTERED_MSNG_BID )
				CASE_RETURN_STRTYPE( MSNG_ERR_UNREGISTERED_MAIN_BID )


				CASE_RETURN_STRTYPE( MSNG_ERR_SYSTEM )
				CASE_RETURN_STRTYPE( MSNG_ERR_OPENDB )
				CASE_RETURN_STRTYPE( MSNG_ERR_SELECTDB )
				CASE_RETURN_STRTYPE( MSNG_ERR_SENDTOWAS )
				CASE_RETURN_STRTYPE( MSNG_ERR_SENDTORS )
				CASE_RETURN_STRTYPE( MSNG_ERR_TIMEOUTWAS )
				CASE_RETURN_STRTYPE( MSNG_ERR_TIMEOUTRS )
				CASE_RETURN_STRTYPE( MSNG_ERR_CLOSEWAS )
				CASE_RETURN_STRTYPE( MSNG_ERR_CLOSERS )
				CASE_RETURN_STRTYPE( MSNG_ERR_RESWAS )
				CASE_RETURN_STRTYPE( MSNG_ERR_RESRS )
				CASE_RETURN_STRTYPE( MSNG_ERR_CLOSEPEER )
				CASE_RETURN_STRTYPE( MSNG_ERR_SENDMESSAGE_FAIL )
				CASE_RETURN_STRTYPE( MSNG_ERR_CRACKED_DATA )

				CASE_RETURN_STRTYPE( MSNG_ERR_RECONNECTNEWJS )
				CASE_RETURN_STRTYPE( MSNG_ERR_UPDATEAUTHKEY )
		}

		return "";
}


const char * CNpMsgrMain::GetServerDefinedCodeMsg(int nCode)
{
	return GetNpIMErrorMsg(nCode);
}

void CNpMsgrMain::Delete()
{
	delete this;
}

void CNpMsgrMain::InitOnlineStatus(E_MSGR_ONLINE_STATUS status)
{
	USER()->m_eOnlineStatus = status;

	// send MSNG_CMD_INITONLINE
	CGBStream stream_send;
	stream_send.CreateStream();
	stream_send.Write_U1( 1 | 2 | 8 );
	stream_send.Write_U1( USER()->m_eOnlineStatus );
	stream_send.Write_U8S_from_Unicode( USER()->Nick() );

	MAINSERVER_SD()->Send_REQUEST_PACKET( static_cast<USHORT>(MSNG_CMD_INITONLINE), 
		static_cast<USHORT>(stream_send.GetDataLen()),
		stream_send.GetData(), NULL );
}


void CNpMsgrMain::SendMessage(unsigned int nUserDefinedMsgCode, const char * pMsg, int nSize, LPCWSTR strToUserID)
{
	CGBStream stream;
	stream.CreateStream(nSize + 100);

	if(strToUserID)
		stream.Write_U8SS_from_Unicode(strToUserID);
	else
		stream.Write_U8SS_from_Unicode(STR_EMPTY_WSTRING);

	stream.Write_U4(nUserDefinedMsgCode);
	stream.Write_U2(nSize);
	stream.Write(pMsg, nSize);

	GetMainServer_SessionDispatcher()->Send_REQUEST_PACKET( static_cast<USHORT>(MSNG_CMD_SEND_MESSAGE), 
		static_cast<USHORT>(stream.GetDataLen()),
		stream.GetData(),
		this );
}


void CNpMsgrMain::ChangeOnlineStatus(E_MSGR_ONLINE_STATUS status, const char *pCustomDataBuffer, int nBufferSize, bool bCustomDataDBSave)
{
	CGBStream stream;
	stream.CreateStream();

	stream.Write_U1( status );
	stream.Write_U1( bCustomDataDBSave == true ? 1 : 0 );

	if(pCustomDataBuffer != NULL && nBufferSize > 0)
	{
		stream.Write_U2(nBufferSize);
		stream.Write(pCustomDataBuffer, nBufferSize);

		m_strCustomData.assign(pCustomDataBuffer, nBufferSize);
	}

	GetMainServer_SessionDispatcher()->Send_REQUEST_PACKET( static_cast<USHORT>(MSNG_CMD_SETSTATUS), 
		static_cast<USHORT>(stream.GetDataLen()),
		stream.GetData(),
		this );

	m_eNewOnlineStatus = status;
}

void CNpMsgrMain::ChangeNick(LPCWSTR wszNick)
{
	CGBStream stream;
	stream.CreateStream();
	
	USES_UTF8_CONVERSION;
	
#if defined(WIN32)
	stream.Write_U8S((_t_UTF8)wszNick);
#else
	stream.Write_U8S_from_Unicode(wszNick);
#endif	

	GetMainServer_SessionDispatcher()->Send_REQUEST_PACKET( static_cast<USHORT>(MSNG_CMD_NOTIFYNICKNAME), 
		static_cast<USHORT>(stream.GetDataLen()),
		stream.GetData(),
		this );

	m_wstrNewNick = wszNick;

	// KEH_API
	//     ʰ ִ.
	//  Ŭ  뺸  ֵ .
	//   Ŷ ֵ  Ǹ ⵵ Ǿ!
	m_pEvents->NpMsgr_OnResult_ChangeNick(this, 0, m_wstrNewNick.c_str());
}

void CNpMsgrMain::RegiFavoriteToPOI(vector<wstring> &vecUsers, USHORT nUserGrp)
{
	CGBStream stream;
	stream.CreateStream(vecUsers.size() * LEN_ID);

	stream.Write_U8SS((_t_UTF8)USER()->Id());

	stream.Write_U2( nUserGrp );

	stream.Write_U1( vecUsers.size() );

	for( size_t i = 0; i < vecUsers.size(); i++)
	{
		stream.Write_U8SS((_t_UTF8)((LPCWSTR)vecUsers[i].c_str()));
	}

	GetMainServer_SessionDispatcher()->Send_REQUEST_PACKET( static_cast<USHORT>(MSNG_CMD_REGIPOI), 
		static_cast<USHORT>(stream.GetDataLen()),
		stream.GetData(),
		this );
}

void CNpMsgrMain::UregFavoriteFromPOI(USHORT nUserGrp)
{
	CGBStream stream;
	stream.CreateStream();

	stream.Write_U8SS((_t_UTF8)USER()->Id());

	stream.Write_U2( nUserGrp );

	GetMainServer_SessionDispatcher()->Send_REQUEST_PACKET( static_cast<USHORT>(MSNG_CMD_UREGPOI), 
		static_cast<USHORT>(stream.GetDataLen()),
		stream.GetData(),
		this );
}

void CNpMsgrMain::UregFavoriteFromPOI(vector<wstring> &vecUsers, USHORT nUserGrp)
{
	CGBStream stream;
	stream.CreateStream(vecUsers.size() * LEN_ID);

	stream.Write_U8SS((_t_UTF8)USER()->Id());

	stream.Write_U2( nUserGrp );

	stream.Write_U1( vecUsers.size() );

	for( size_t i = 0; i < vecUsers.size(); i++)
	{
		stream.Write_U8SS((_t_UTF8)((LPCWSTR)vecUsers[i].c_str()));
	}

	GetMainServer_SessionDispatcher()->Send_REQUEST_PACKET( static_cast<USHORT>(MSNG_CMD_UREGPOI), 
		static_cast<USHORT>(stream.GetDataLen()),
		stream.GetData(),
		this );
}

void CNpMsgrMain::GetBuddyStatus(std::vector<wstring> &vecUsers)
{
	CGBStream stream;
	stream.CreateStream(vecUsers.size() * LEN_ID + sizeof(uint8_t));

	stream.Write_U1( vecUsers.size() );

	for( size_t i = 0; i < vecUsers.size(); i++)
	{
		stream.Write_U8SS((_t_UTF8)((LPCWSTR)vecUsers[i].c_str()));
	}

	GetMainServer_SessionDispatcher()->Send_REQUEST_PACKET( static_cast<USHORT>(MSNG_CMD_GETSTATUS),
		static_cast<USHORT>(stream.GetDataLen()),
		stream.GetData(),
		this );
}


//class _INpSessionDispatcher_PacketHandler
void CALLBACK CNpMsgrMain::INpSessionDispatcher_OnReceive_REQUEST_PACKET(INpSessionDispatcher * pEventSource, const INpPacket * pPacket)
{
	CGBStream stream;
	stream.UseExternalBuffer( pPacket->BodyData(), pPacket->BodyLength() );

	switch( pPacket->M6V2()->usCMMD )
	{
	case MSNG_CMD_SYSTEMNOTIFY: //(0x0700)
		 try
		{
			uint16_t nOsType = stream.Read_U2();

			char *pSystemNotify;
			stream.Read_S( &pSystemNotify );

			m_pEvents->NpMsgr_OnReceive_SystemNotify( pSystemNotify );
		}
		catch( CGBStreamException e )
		{
//			DLOG("receive MSNG_CMD_SYSTEMNOTIFY, parse error!!,");
		}
		break;

	case MSNG_CMD_NOTIFYNOTE:
		try
		{

			OutputDebugString(_T("~~~~~~~~~ MSNG_CMD_NOTIFYNOTE ~~~~~~~~~\r\n"));

			USES_CONVERSION;
			ULONGLONG lNoteId;
			wstring strFromID, strFromName;
			ULONG nUnreadNoteCnt;

			stream.Read_U8(lNoteId);
			stream.Read_U8S_to_Unicode_stl(strFromID);
			stream.Read_U8S_to_Unicode_stl(strFromName);
			nUnreadNoteCnt = stream.Read_U4();

			m_pEvents->NpMsgr_OnReceive_NoteNotify(lNoteId, strFromID.c_str(), strFromName.c_str(), nUnreadNoteCnt);

		}
		catch(CGBStreamException e)
		{
//			DLOG("receive MSNG_CMD_NOTIFYNOTE, parse error!!,");
		}
		break;

	case MSNG_CMD_NOTIFYUNREADNOTECNT:
		try
		{
			OutputDebugString(_T("~~~~~~~~~ MSNG_CMD_NOTIFYUNREADNOTECNT ~~~~~~~~~\r\n"));

			USES_CONVERSION;
			ULONG lUnreadCnt;
			lUnreadCnt = stream.Read_U4();			
			m_pEvents->NpMsgr_OnReceive_NoteUnreadCntNotify(lUnreadCnt);

		}
		catch(CGBStreamException e)
		{
//			DLOG("receive MSNG_CMD_NOTIFYUNREADNOTECNT, parse error!!,");
		}
		break;

	case MSNG_CMD_NOTIFYMEMO:
		try
		{
			USES_CONVERSION;

			WSTRING wszRegisterID, wszTitle, wszContents;
			BYTE nRegisterIDLen, nMsgType;
			ULONG ulTime;
			USHORT usTitleLen, ulContentsLen;

			nRegisterIDLen = (BYTE)stream.Read_U8SS_to_Unicode_stl(wszRegisterID);			
			ulTime = stream.Read_U4();			
			nMsgType = stream.Read_U1(); //reserved			
			usTitleLen = stream.Read_U8S_to_Unicode_stl(wszTitle);			
			ulContentsLen = stream.Read_U8S_to_Unicode_stl(wszContents);			

			m_pEvents->NpMsgr_OnReceive_MemoNotify(wszRegisterID.c_str(), ulTime, 1/*Unicode*/, wszTitle.c_str(), 
				(void*)wszContents.c_str(), ulContentsLen, 0);

		}
		catch(CGBStreamException e)
		{
//			DLOG("receive MSNG_CMD_NOTIFYMEMO, parse error!!,");
		}
		break;
	}
}



void CALLBACK CNpMsgrMain::INpSessionDispatcher_OnReceive_RESPONSE_PACKET(INpSessionDispatcher * pEventSource, const INpPacket * pOriginalRequestPacket, const INpPacket * pPacket)
{
	int result = 0xFFFFFFFF;
	USHORT usResultCode = MSNG_ERR_SYSTEM;

	CGBStream stream;
	stream.UseExternalBuffer( pPacket->BodyData(), pPacket->BodyLength() );

	switch(pPacket->M6V2()->usCMMD)
	{
	case GET_RESPONSE_CMMD( MSNG_CMD_GETSTATUS ):
		{
			GetMainServer_SessionDispatcher()->Deregister_RESPONSE_PACKET_Handler( pPacket->M6V2()->ulTrId );

			uint8_t nUserCnt = stream.Read_U1();

			std::map<WSTRING, BYTE> mapBuddyInfo;
			for( size_t i = 0; i < nUserCnt; i++ )
			{
				WSTRING strUserID;
				stream.Read_U8SS_to_Unicode_stl(strUserID);

				uint8_t online_status = stream.Read_U1();

				if( online_status == ST_OFFLINE)
					continue;

				std::pair<std::map<WSTRING, BYTE>::iterator, bool> ret = mapBuddyInfo.insert(std::map<WSTRING, BYTE>::value_type(strUserID, online_status));
			}

			m_pEvents->NpMsgr_OnResult_BuddyInfo_ActiveXChat(this, mapBuddyInfo);
		}
		break;
	case GET_RESPONSE_CMMD( MSNG_CMD_SEND_MESSAGE ):
		{
			GetMainServer_SessionDispatcher()->Deregister_RESPONSE_PACKET_Handler( pPacket->M6V2()->ulTrId );
			try
			{
				usResultCode = stream.Read_U2();
			}
			catch( CGBStreamException e )
			{
			}

			m_pEvents->NpMsgr_OnResult_SendMessage(this, usResultCode);
		}
		break;
	case GET_RESPONSE_CMMD( MSNG_CMD_SETSTATUS ):
		{
			GetMainServer_SessionDispatcher()->Deregister_RESPONSE_PACKET_Handler( pPacket->M6V2()->ulTrId );

			try
			{
				usResultCode = stream.Read_U2();
				if( usResultCode == S_OK )
				{
					result = 0;
					USER()->m_eOnlineStatus = m_eNewOnlineStatus;
				}
			}
			catch( CGBStreamException e )
			{
			}

			m_pEvents->NpMsgr_OnResult_ChangeOnlineStatus(this, result, m_eNewOnlineStatus);
		}
		break;

	case GET_RESPONSE_CMMD( MSNG_CMD_NOTIFYNICKNAME ):
		{
			GetMainServer_SessionDispatcher()->Deregister_RESPONSE_PACKET_Handler( pPacket->M6V2()->ulTrId );

			try
			{
				usResultCode = stream.Read_U2();
				if( usResultCode == S_OK )
				{
					result = 0;
					USER()->m_wstrNick = m_wstrNewNick;
				}
			}
			catch( CGBStreamException e )
			{
			}

			m_pEvents->NpMsgr_OnResult_ChangeNick(this, result, m_wstrNewNick.c_str());
		}
		break;
	case RESPONSE(MSNG_CMD_CALLWAS):
		try
		{
			GetMainServer_SessionDispatcher()->Deregister_RESPONSE_PACKET_Handler( pPacket->M6V2()->ulTrId );

			USHORT usResultCode = stream.Read_U2();

			if( usResultCode == 200 )
			{
				string strResponse;
				stream.Read_S_stl( strResponse );

				if( strnicmp( strResponse.c_str(), "OK:", 3 ) == 0 )
				{
					m_pNpMsgrUser->SetAuthCookie(strResponse.c_str() + 3);
				}
				else
				{
				}
			}
			else
			{
			}
		}
		catch( CGBStreamException e )
		{
		}
		break;
	}
}


void CALLBACK CNpMsgrMain::INpSessionDispatcher_OnReceive_CToC(INpSessionDispatcher * pEventSource, USHORT usCToC_SubCMMD, const INpPacket * pPacket)
{
}



void CNpMsgrMain::ReadBuddyDataCacheSerial(ULONG &nBuddyListCacheSerial, ULONG &nGroupListCacheSerial )
{
	size_t dwLength = 0;
	char * pData = NULL;
	char szFilePath[MAX_PATH+1];

	nBuddyListCacheSerial = 0;
	nGroupListCacheSerial = 0;

#if defined(WIN32)	
	// ϴ cache  .
#ifndef _DEBUG
	return;
#endif
#endif

	// buddy cache
	CONFIG()->GetFilePath( "BuddyListCache", szFilePath );
	pData = GTCommonAPI::ReadDataFileAlloc( szFilePath, dwLength );
	if( pData )
	{
		if( dwLength > 4 )
			memcpy( &nBuddyListCacheSerial, pData, 4 );
		delete [] pData;
	}

	// group cache
	CONFIG()->GetFilePath( "GroupListCache", szFilePath );
	pData = GTCommonAPI::ReadDataFileAlloc( szFilePath, dwLength );
	if( pData )
	{
		if( dwLength > 4 )
			memcpy( &nGroupListCacheSerial, pData, 4 );
		delete [] pData;
	}
}


bool CNpMsgrMain::OnTimerProcess(WPARAM wParam, LPARAM lParam)
{
	if( wParam == IDT_RELOGIN_START )
	{
		TIMER_KILL(IDT_RELOGIN_START );
		
//		DLOG(">> CNpMsgrMain::OnTimerProcess : %d %d", USER()->OnlineStatus(), USER()->m_eLoginStatus);
		if( USER()->m_eLoginStatus == LS_RELOGIN_WAIT_START )
		{
			USER()->m_eLoginStatus = LS_RELOGIN_CONNECTING;
			if(Internal_LoginConnect() == false)
			{
				OnInternalReloginResult( RC_CONNECTFAIL, -1 );
			}
		}
	}
	else if (wParam == IDT_CHECK_NETWORK)
	{
		//USES_CONVERSION;
		//bool bConnected = CheckConnection(_T("biz-ntalk.ncloudbiz.com"), 80, 3);

		//if( bConnected == false )
		//{
		//	if( m_bPrevNetworkCheckIsFalse )
		//	{
		//		GT_FLOG(L"IDT_CHECK_NETWORK [DISCONNECTED]");
		//		OutputDebugString(_T("\r\n====================== IDT_CHECK_NETWORK [DISCONNECTED] ======================\r\n"));

		//		m_pEvents->NpMsgr_OnDisconnected( this, 1 );
		//		m_pMainServerNpSession->CloseSession();
		//		OnInternalMainSessionClosed(0);

		//		m_bPrevNetworkCheckIsFalse = false;
		//	}
		//	else
		//		m_bPrevNetworkCheckIsFalse = true;
		//}
		//else
		//	m_bPrevNetworkCheckIsFalse = false;

		m_pEvents->NpMsgr_CheckClientNetwork();
	}

	return true;
}

bool CNpMsgrMain::OnSignalProcess(int nMsg, WPARAM wParam, LPARAM lParam)
{
	return true;
}


// class _CIkWndHandler
void CALLBACK CNpMsgrMain::IkWnd_OnMessage(CIkWnd * pIkWnd, UINT message, WPARAM wParam, LPARAM lParam )
{
	switch( message )
	{
	case WM_TIMER:
		OnTimerProcess(wParam, lParam);
		break;
	}
}


////////////////////////////////////////////////////////////////////////////
// class _INpSessionEvents
bool ReadRegistryData(LPCTSTR szDataName, LPBYTE pOutBuff, int * pcbBuffSize, LPDWORD lpType,
					  LPCTSTR szSection, LPCTSTR szAppName = _T("Game Talk"), LPCTSTR szCompany = _T("NHN Corporation") )
{
	TCHAR KeyStr[1024];
	HKEY hRegKey = NULL;

	int nBuffSize = *pcbBuffSize;	
	memset(pOutBuff, 0, nBuffSize);

	_sntprintf(KeyStr, 1024, _T("Software\\%s\\%s\\%s"), szCompany, szAppName, szSection);
	if(RegOpenKeyEx(HKEY_LOCAL_MACHINE, KeyStr, 0, KEY_READ, &hRegKey) != ERROR_SUCCESS)
	{
		*pcbBuffSize = 0;
		return false;
	}

	LONG lResult = RegQueryValueEx(hRegKey, szDataName, 0, lpType, pOutBuff, (LPDWORD)pcbBuffSize);

	RegCloseKey(hRegKey);

	ASSERT( nBuffSize >= *pcbBuffSize );

	if( ERROR_MORE_DATA == lResult )
	{
		ASSERT( nBuffSize < *pcbBuffSize );
		return false;
	}
	else if( ERROR_SUCCESS != lResult )
	{
		*pcbBuffSize = 0;
		return false;
	}

	return true;
}

bool WriteRegistryData(LPCTSTR szDataName, LPBYTE pData, int nDataLen, bool bBinary,
					   LPCTSTR szSection, LPCTSTR szAppName = _T("Game Talk"), LPCTSTR szCompany = _T("NHN Corporation") )
{
	TCHAR szKeyStr[1024];
	HKEY hRegKey = NULL;

	_sntprintf(szKeyStr, 1024, _T("Software\\%s\\%s\\%s"), szCompany, szAppName, szSection);

	//LONG lResult = RegOpenKeyEx(HKEY_LOCAL_MACHINE, szKeyStr, 0, KEY_WRITE, &hRegKey);
	LONG lResult;	
	DWORD dwDisposition;

	lResult = RegCreateKeyEx(HKEY_LOCAL_MACHINE, szKeyStr, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hRegKey, &dwDisposition);
	if( lResult != ERROR_SUCCESS )
	{
		return false;
	}

//	IK_TRACE("dwDisposition == %s\n", dwDisposition == REG_CREATED_NEW_KEY ? "REG_CREATED_NEW_KEY" : "REG_OPENED_EXISTING_KEY" );

	lResult = RegSetValueEx(hRegKey, szDataName, 0, bBinary ? REG_BINARY : REG_SZ, pData, nDataLen );

	RegCloseKey(hRegKey);

	if( ERROR_MORE_DATA == lResult )
	{
		return false;
	}
	else if( ERROR_SUCCESS != lResult )
	{
		int nError = GetLastError();

		return false;
	}

	return true;
}


void CALLBACK CNpMsgrMain::OnNpSession_Connect(INpSession *pSession, int nSocketError)
{
	nfc::AutoLock<nfc::CMutex> autolock(&m_mutexSession);
	
	if( pSession == m_pLoginNpSession )
	{
		switch( USER()->m_eLoginStatus )
		{
		case LS_LOGIN_CONNECTING:
			if( nSocketError )
				OnInternalLoginResult( RC_CONNECTFAIL, nSocketError );
			else
				Send_LOGIN();
			break;

		case LS_RELOGIN_CONNECTING:
			if( nSocketError )
				OnInternalReloginResult( RC_CONNECTFAIL, nSocketError );
			else
				Send_RELOGIN();
			break;

		default:
			USER()->m_eLoginStatus = LS_LOGIN_READY;
			USER()->m_eOnlineStatus = ST_OFFLINE;

			m_pEvents->NpMsgr_SetLoginWorking( this, false );

			// ̷ Ȳ ߻ϸ ȵ.
			ASSERT( false );
			pSession->CloseSession();
			break;
		}
	}
	else
	{
		USER()->m_eLoginStatus = LS_LOGIN_READY;
		USER()->m_eOnlineStatus = ST_OFFLINE;

		m_pEvents->NpMsgr_SetLoginWorking( this, false );

		// ̷ Ȳ ߻ϸ ȵ.
		ASSERT( false );
		pSession->CloseSession();
	}	
}

#define BASE64_FLAG_NONE	0
#define BASE64_FLAG_NOPAD	1
#define BASE64_FLAG_NOCRLF  2

inline int Base64EncodeGetRequiredLength(int nSrcLen, DWORD dwFlags=BASE64_FLAG_NONE) throw()
{
	int nRet = nSrcLen*4/3;

	if ((dwFlags & BASE64_FLAG_NOPAD) == 0)
		nRet += nSrcLen % 3;

	int nCRLFs = nRet / 76 + 1;
	int nOnLastLine = nRet % 76;

	if (nOnLastLine)
	{
		if (nOnLastLine % 4)
			nRet += 4-(nOnLastLine % 4);
	}

	nCRLFs *= 2;

	if ((dwFlags & BASE64_FLAG_NOCRLF) == 0)
		nRet += nCRLFs;

	return nRet;
}

inline int Base64DecodeGetRequiredLength(int nSrcLen) throw()
{
	return nSrcLen;
}

inline bool Base64Encode(
	const BYTE *pbSrcData,
	int nSrcLen,
	LPSTR szDest,
	int *pnDestLen,
	DWORD dwFlags=BASE64_FLAG_NONE) throw()
{
	static const char s_chBase64EncodingTable[64] = {
		'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q',
		'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f', 'g',	'h',
		'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y',
		'z', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+', '/' };

	if (!pbSrcData || !szDest || !pnDestLen)
	{
		return false;
	}

	int nWritten( 0 );
	int nLen1( (nSrcLen/3)*4 );
	int nLen2( nLen1/76 );
	int nLen3( 19 );

	for (int i=0; i<=nLen2; i++)
	{
		if (i==nLen2)
			nLen3 = (nLen1%76)/4;

		for (int j=0; j<nLen3; j++)
		{
			DWORD dwCurr(0);
			for (int n=0; n<3; n++)
			{
				dwCurr |= *pbSrcData++;
				dwCurr <<= 8;
			}
			for (int k=0; k<4; k++)
			{
				BYTE b = (BYTE)(dwCurr>>26);
				*szDest++ = s_chBase64EncodingTable[b];
				dwCurr <<= 6;
			}
		}
		nWritten+= nLen3*4;

		if ((dwFlags & BASE64_FLAG_NOCRLF)==0)
		{
			*szDest++ = '\r';
			*szDest++ = '\n';
			nWritten+= 2;
		}
	}

	if (nWritten && (dwFlags & BASE64_FLAG_NOCRLF)==0)
	{
		szDest-= 2;
		nWritten -= 2;
	}

	nLen2 = nSrcLen%3 ? nSrcLen%3 + 1 : 0;
	if (nLen2)
	{
		DWORD dwCurr(0);
		for (int n=0; n<3; n++)
		{
			if (n<(nSrcLen%3))
				dwCurr |= *pbSrcData++;
			dwCurr <<= 8;
		}
		for (int k=0; k<nLen2; k++)
		{
			BYTE b = (BYTE)(dwCurr>>26);
			*szDest++ = s_chBase64EncodingTable[b];
			dwCurr <<= 6;
		}
		nWritten+= nLen2;
		if ((dwFlags & BASE64_FLAG_NOPAD)==0)
		{
			nLen3 = nLen2 ? 4-nLen2 : 0;
			for (int j=0; j<nLen3; j++)
			{
				*szDest++ = '=';
			}
			nWritten+= nLen3;
		}
	}

	*pnDestLen = nWritten;
	return true;
}


//-----------------------------------------------------------------------------------------------------------
// !
// ݵ userid seed ȣȭҶ! userid   ID̾ Ѵ! eunsiri00@hangame.com
//-----------------------------------------------------------------------------------------------------------

void CNpMsgrMain::GetEncryptUserPassword(BYTE &bytePwdCryptType, BYTE biEncryptedPwd[], USHORT &usEncryptedPwdLength)
{
	int nPwdCryptKeyBaseSerial = 0;

	EncryptString(USER()->m_strPassword, (const char*)(_t_UTF8)USER()->Id(), bytePwdCryptType, nPwdCryptKeyBaseSerial, biEncryptedPwd, usEncryptedPwdLength );

}

void CNpMsgrMain::EncryptString(LPCSTR szData, LPCSTR szSeed, BYTE &bytePwdCryptType, int &nPwdCryptKeyBaseSerial, BYTE biEncryptedPwd[], USHORT &usEncryptedPwdLength)
{
	string strData(szData);
	string strSeed(szSeed);
	
	EncryptString(strData, strSeed, bytePwdCryptType, nPwdCryptKeyBaseSerial, biEncryptedPwd, usEncryptedPwdLength);
}


#define ENCRYPT_UTF8_BIT	0x80

void CNpMsgrMain::EncryptString(const string& strData, const string& strSeed, BYTE &bytePwdCryptType, int &nPwdCryptKeyBaseSerial, BYTE biEncryptedPwd[], USHORT &usEncryptedPwdLength)
{
	char szKeyBase[512] = {0};

	//bytePwdCryptType = 0;
	bytePwdCryptType = 2;

	if( bytePwdCryptType == 0 )
	{
		nPwdCryptKeyBaseSerial = 0;

		usEncryptedPwdLength = strData.length();
		memcpy( biEncryptedPwd, strData.c_str(), usEncryptedPwdLength );
	}
	else if( bytePwdCryptType == 2 )
	{
		nPwdCryptKeyBaseSerial = 0;
#if defined(WIN32)	// __KEH_USE_OPENSSL__
		char szKeyBaseSerial[32] = {0};
		int nOutputLength = 32;
		if( ReadRegistryData("PWD Crypt Key Base Serial", (LPBYTE)szKeyBaseSerial, &nOutputLength, NULL, "Info") )
		{
			nPwdCryptKeyBaseSerial = atoi( szKeyBaseSerial );
			if( nPwdCryptKeyBaseSerial > 0 )
			{
				nOutputLength = 512;
				if( ReadRegistryData("PWD Crypt Key Base", (LPBYTE)szKeyBase, &nOutputLength, NULL, "Info") )
				{
					if( nOutputLength == 0 || strlen( szKeyBase ) == 0 )
						nPwdCryptKeyBaseSerial = 0;
				}
			}
		}
#endif
		if( nPwdCryptKeyBaseSerial == 0 )
		{
			char szDefaultKeyTable[] = 
				"0ZO67A8d60SM0uZvZlU347DGHmVt9505pxh447U2gx8BNO342r6gin4STd9t9x7A2qHI8ukxDyw9nG5o2491xg82w707MMrdS1Zrn8QAP3Za23F4E8HsxVWTyNuFPmFNPc2dlInGyBR6w2s80Nfx16";
			
#if defined(WIN32)	// __KEH_USE_OPENSSL__
			ASSERT( sizeof( szKeyBase ) > strlen(szDefaultKeyTable) );
#endif
			strcpy( szKeyBase, szDefaultKeyTable );
		}
		
#if defined(__APPLE_CC__)
		bytePwdCryptType = ENCRYPT_UTF8_BIT | 2;
#endif
		
		// base64 encode for korean id
		std::string id = strSeed;
		std::vector<char> enc64ID(id.length()*2); // prepare the buffer
		int encSize = 0;
		Base64Encode((const BYTE*)id.c_str(), id.length(), &enc64ID[0], &encSize); // this is no return value for this func ^^;

#ifdef _DEBUG
		{
			// test
			std::ofstream file("id_log.txt");
			file << id << std::endl;
			file << &enc64ID[0] << std::endl;
		}
#endif
		
		int nEncryptedPwd_len = usEncryptedPwdLength;
		pwdcrypt_encrypt(strData.c_str(), strData.length(),
			biEncryptedPwd, &nEncryptedPwd_len,
			&enc64ID[0],
			szKeyBase, 
			strlen( szKeyBase ) );

		usEncryptedPwdLength = nEncryptedPwd_len;
	}
}


#if defined(WIN32)
typedef DWORD (__stdcall * VERSIZEPROC)(LPCTSTR, LPDWORD);
typedef BOOL (__stdcall *  VERINFOPROC)(LPCTSTR, DWORD, DWORD, LPVOID);
typedef BOOL (__stdcall *  VERQUERYPROC)(const LPVOID, LPTSTR, LPVOID, PUINT);

string GetFileVersionStr(LPCSTR szFileName)
{
	string strVersionStr;

	VERSIZEPROC GetFileVersionInfoSize = NULL;
	VERINFOPROC GetFileVersionInfo = NULL;
	VERQUERYPROC VerQueryValue = NULL;

	HMODULE hLib = LoadLibrary("Version.dll");
	if(!hLib)
		return strVersionStr;

	GetFileVersionInfoSize = (VERSIZEPROC)GetProcAddress(hLib, "GetFileVersionInfoSizeA");
	GetFileVersionInfo = (VERINFOPROC)GetProcAddress(hLib, "GetFileVersionInfoA");
	VerQueryValue = (VERQUERYPROC)GetProcAddress(hLib, "VerQueryValueA");

	if(GetFileVersionInfoSize != NULL)
	{
		int nSize = GetFileVersionInfoSize(szFileName, NULL);

		char *pBuffer = new char[nSize];
		if(NULL == pBuffer)
			return strVersionStr;

		if( GetFileVersionInfo(szFileName, NULL, nSize, (LPVOID)pBuffer) )
		{
			unsigned int len;
			VS_FIXEDFILEINFO * ver;
			VerQueryValue(pBuffer, "\\", (void **)&ver, &len);

			char szTmp[256];
			_snprintf(szTmp, 256, "%d.%d.%d.%d", //, date:0x%08X-0x%08X", 
				HIWORD(ver->dwProductVersionMS),
				LOWORD(ver->dwProductVersionMS),
				HIWORD(ver->dwProductVersionLS),
				LOWORD(ver->dwProductVersionLS)); //,

			strVersionStr = szTmp;
		}

		delete []pBuffer;
	}
	FreeLibrary(hLib);

	return strVersionStr;
}

string GetVersionStr()
{
	string strVersion;

	char szFileName[MAX_PATH];

	GetModuleFileName(NULL, szFileName, MAX_PATH);
/*	char * pPureFilename = strrchr(szFileName, '\\');
	if( pPureFilename )
	{
		pPureFilename++;
		strVersion = pPureFilename;
	}
*/
	// PC  NtalkPC ؾ   .
	strVersion = "NtalkPC," + GetFileVersionStr( szFileName );

	return strVersion;
}


string GetFileSizeStr()
{
	string strSize;

	char szFileName[MAX_PATH];

	GetModuleFileName(NULL, szFileName, MAX_PATH);

	HANDLE hFile = CreateFile(szFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if(hFile == INVALID_HANDLE_VALUE)
	{
		return "Size:Unknown";
	}

	DWORD dwFileSize = ::GetFileSize(hFile, NULL);
	CloseHandle(hFile);

	char szTmp[256];
	_snprintf(szTmp, 255, "Size:%d", dwFileSize);
	strSize = szTmp;
	return strSize;
}


string GetOsVersion()
{
	OSVERSIONINFOEX osvi;
	BOOL bOsVersionInfoEx;

	string Verstr;
#define   BUFSIZE 80

	// Try calling GetVersionEx using the OSVERSIONINFOEX structure.
	// If that fails, try using the OSVERSIONINFO structure.

	ZeroMemory(&osvi, sizeof(OSVERSIONINFOEX));
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);

	if( !(bOsVersionInfoEx = GetVersionEx ((OSVERSIONINFO *) &osvi)) )
	{
		osvi.dwOSVersionInfoSize = sizeof (OSVERSIONINFO);
		if (! GetVersionEx ( (OSVERSIONINFO *) &osvi) ) 
			return FALSE;
	}

	switch (osvi.dwPlatformId)
	{
		// Test for the Windows NT product family.
	case VER_PLATFORM_WIN32_NT:

		// Test for the specific product family.

		if ( osvi.dwMajorVersion == 6 && osvi.dwMinorVersion == 0 )
		{
			if( osvi.wProductType == VER_NT_WORKSTATION )
				Verstr += "Windows Vista ";
			else Verstr += "Windows Server Longhorn " ;
		}
		else if (osvi.dwMajorVersion >= 6)
		{
			char strMsg[80];
			sprintf(strMsg, "Windows %d.%d", osvi.dwMajorVersion, osvi.dwMinorVersion);
			Verstr+=strMsg;
		}

		if ( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 2 )
			Verstr += "Microsoft Windows Server 2003 family, ";

		if ( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 1 )
			Verstr += "Microsoft Windows XP ";

		if ( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 0 )
			Verstr += "Microsoft Windows 2000 ";

		if ( osvi.dwMajorVersion <= 4 )
			Verstr += "Microsoft Windows NT ";

		// Test for specific product on Windows NT 4.0 SP6 and later.
		if( bOsVersionInfoEx )
		{
			// Test for the workstation type.
			if ( osvi.wProductType == VER_NT_WORKSTATION )
			{
				if( osvi.dwMajorVersion == 4 )
					Verstr +=  "Workstation 4.0 ";
				else if( osvi.wSuiteMask & VER_SUITE_PERSONAL )
					Verstr +=  "Home Edition ";
				else
					Verstr +=  "Professional ";
			}

			// Test for the server type.
			else if ( osvi.wProductType == VER_NT_SERVER )
			{
				if( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 2 )
				{
					if( osvi.wSuiteMask & VER_SUITE_DATACENTER )
						Verstr +=  "Datacenter Edition " ;
					else if( osvi.wSuiteMask & VER_SUITE_ENTERPRISE )
						Verstr +=  "Enterprise Edition " ;
					else if ( osvi.wSuiteMask == VER_SUITE_BLADE )
						Verstr +=  "Web Edition " ;
					else
						Verstr +=  "Standard Edition " ;
				}

				else if( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 0 )
				{
					if( osvi.wSuiteMask & VER_SUITE_DATACENTER )
						Verstr +=  "Datacenter Server " ;
					else if( osvi.wSuiteMask & VER_SUITE_ENTERPRISE )
						Verstr +=  "Advanced Server " ;
					else
						Verstr +=  "Server " ;
				}

				else  // Windows NT 4.0 
				{
					if( osvi.wSuiteMask & VER_SUITE_ENTERPRISE )
						Verstr += "Server 4.0, Enterprise Edition " ;
					else
						Verstr +=  "Server 4.0 " ;
				}
			}
		}
		else  // Test for specific product on Windows NT 4.0 SP5 and earlier
		{
			HKEY hKey;
			char szProductType[BUFSIZE];
			DWORD dwBufLen=BUFSIZE;
			LONG lRet;

			lRet = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
				"SYSTEM\\CurrentControlSet\\Control\\ProductOptions",
				0, KEY_QUERY_VALUE, &hKey );
			if( lRet != ERROR_SUCCESS )
				return FALSE;

			lRet = RegQueryValueEx( hKey, "ProductType", NULL, NULL,
				(LPBYTE) szProductType, &dwBufLen);
			if( (lRet != ERROR_SUCCESS) || (dwBufLen > BUFSIZE) )
				return FALSE;

			RegCloseKey( hKey );

			if ( lstrcmpi( "WINNT", szProductType) == 0 )
				Verstr +=  "Workstation " ;
			if ( lstrcmpi( "LANMANNT", szProductType) == 0 )
				Verstr +=  "Server " ;
			if ( lstrcmpi( "SERVERNT", szProductType) == 0 )
				Verstr +=  "Advanced Server " ;

			//printf( "%d.%d ", osvi.dwMajorVersion, osvi.dwMinorVersion );
		}

		// Display service pack (if any) and build number.

		if( osvi.dwMajorVersion == 4 && 
			lstrcmpi( osvi.szCSDVersion, "Service Pack 6" ) == 0 )
		{
			HKEY hKey;
			LONG lRet;

			// Test for SP6 versus SP6a.
			lRet = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
				"SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Hotfix\\Q246009",
				0, KEY_QUERY_VALUE, &hKey );
			if( lRet == ERROR_SUCCESS )
				Verstr +=  "Service Pack 6a";//, osvi.dwBuildNumber & 0xFFFF );         
			else // Windows NT 4.0 prior to SP6a
			{

				//printf( "%s (Build %d)\n",
				Verstr +=   osvi.szCSDVersion;
				// osvi.dwBuildNumber & 0xFFFF);
			}

			RegCloseKey( hKey );
		}
		else // Windows NT 3.51 and earlier or Windows 2000 and later
		{
			Verstr += osvi.szCSDVersion;
			//printf( "%s (Build %d)\n",
			// osvi.szCSDVersion,
			//osvi.dwBuildNumber & 0xFFFF);
		}


		break;

		// Test for the Windows 95 product family.
	case VER_PLATFORM_WIN32_WINDOWS:

		if (osvi.dwMajorVersion == 4 && osvi.dwMinorVersion == 0)
		{
			Verstr += "Microsoft Windows 95 ";
			if ( osvi.szCSDVersion[1] == 'C' || osvi.szCSDVersion[1] == 'B' )
				Verstr += "OSR2 " ;
		} 

		if (osvi.dwMajorVersion == 4 && osvi.dwMinorVersion == 10)
		{
			Verstr += "Microsoft Windows 98 ";
			if ( osvi.szCSDVersion[1] == 'A' )
				Verstr += "SE " ;
		} 

		if (osvi.dwMajorVersion == 4 && osvi.dwMinorVersion == 90)
		{
			Verstr += "Microsoft Windows Millennium Edition";
		} 
		break;

	case VER_PLATFORM_WIN32s:

		Verstr += "Microsoft Win32s";
		break;
	}
	return Verstr;
}
#endif // WIN32


void CNpMsgrMain::Send_LOGIN()
{
	// send LOGIN
	const char *pp = USER()->AuthCookie();
	const UTF8 * pszUTF8Ticket= (const UTF8 *)USER()->AuthCookie();
	char szUrlTicket[2048];
	_snprintf(szUrlTicket, 2048, "%s", pszUTF8Ticket);


	BYTE bytePwdCryptType;
	int nPwdCryptKeyBaseSerial = 0;
	BYTE biEncryptedPwd[256];
	USHORT usEncryptedPwdLength = 256;
	USHORT usEncryptedMobileTokenLength = 256;
	USHORT usEncryptedMobileNumberLength = 256;

	EncryptString(USER()->m_strPassword, (const char*)(_t_UTF8)USER()->Id(), bytePwdCryptType, nPwdCryptKeyBaseSerial, biEncryptedPwd, usEncryptedPwdLength );

	string strVersionStr = GetVersionStr();
	strVersionStr += "|";
	strVersionStr += GetOsVersion();

	// make LOGIN packet
	CGBStream stream;
	stream.CreateStream();


	try
	{
		// User ID
		stream.Write_U8SS((_t_UTF8)USER()->Id());		

		stream.Write( &bytePwdCryptType, 1 );					// enc type
		stream.Write( &usEncryptedPwdLength, 2 );				// encrypted data length
		stream.Write( biEncryptedPwd, usEncryptedPwdLength );	// encrypted data

		stream.Write_U4(0);	// password key serial

		// OS Type
		stream.Write_U2(m_usOSType);	// ostype

		// OS Version
		stream.Write_SS( strVersionStr );

		m_pLoginNpSession->SendNpPacket_M6V2( static_cast<USHORT>(MSNG_CMD_LOGIN), 0, static_cast<USHORT>(stream.GetDataLen()), stream.GetData() );

		USER()->m_eLoginStatus = LS_LOGIN_SEND_DATA;
	}
	catch( CGBStreamException e )
	{
		OnInternalLoginResult( RC_GBSTREAM_MAKE_ERROR, e.Error() );
	}
}


void CNpMsgrMain::Send_RELOGIN()
{
	// send RELOGIN

	BYTE bytePwdCryptType;
	int nPwdCryptKeyBaseSerial = 0;
	BYTE biEncryptedPwd[256];
	USHORT usEncryptedPwdLength = 256;
	
	// You may need to get new ticket.. (SILOS auth ticket must be fresh status..)
	m_pEvents->NpMsgr_NeedUpdateLoginTicket();

	const UTF8 * pszUTF8Ticket= (const UTF8 *)USER()->AuthCookie();
	char szUrlTicket[2048];
	USHORT usEncryptedMobileNumberLength = 256;
	
	_snprintf(szUrlTicket, 2048, "%s", pszUTF8Ticket);

	EncryptString(USER()->m_strPassword, (const char*)(_t_UTF8)USER()->LoginInputId(), bytePwdCryptType, nPwdCryptKeyBaseSerial, biEncryptedPwd, usEncryptedPwdLength );

	string strVersionStr = GetVersionStr();
	strVersionStr += "|";
	strVersionStr += GetOsVersion();

	// make RELOGIN packet
	CGBStream stream;
	stream.CreateStream();

	try
	{
		/*
		SS	A ޽ ̵.
			KS	A н
			U2	 ڵ
			U4	KEY ĳ .
			*/
		
		stream.Write_U8SS( (_t_UTF8)USER()->LoginInputId() );
		
		stream.Write( &bytePwdCryptType, 1 );					// enc type
		stream.Write( &usEncryptedPwdLength, 2 );				// encrypted data length
		stream.Write( biEncryptedPwd, usEncryptedPwdLength );	// encrypted data
		
		stream.Write_U4( 0 );
		
		// OS Type
		stream.Write_U2(m_usOSType);

		// OS Version
		stream.Write_SS( strVersionStr );
	}
	catch( CGBStreamException e )
	{
		OnInternalReloginResult( RC_GBSTREAM_MAKE_ERROR, e.Error() );
		return;
	}

	OutputDebugString(_T("\r\n##### CNpMsgrMain::Send_RELOGIN #####"));

	m_pLoginNpSession->SendNpPacket_M6V2( static_cast<USHORT>(MSNG_CMD_RELOGIN), 0, 
		static_cast<USHORT>(stream.GetDataLen()), stream.GetData() );
	USER()->m_eLoginStatus = LS_RELOGIN_SEND_DATA;
}


void CALLBACK CNpMsgrMain::OnNpSession_Close(INpSession *pSession, int nSocketError, bool bRecvInvalidPacket)
{
	if( m_pLoginNpSession == pSession )
	{
		switch( USER()->m_eLoginStatus )
		{
		case LS_LOGIN_READY:
			ASSERT( false );
			pSession->CloseSession();
			break;

		case LS_LOGIN_CONNECTING:
		case LS_LOGIN_SEND_DATA:
		case LS_LOGIN_RECV_DATA:
			OnInternalLoginResult( RC_SOCKET_CLOSED, nSocketError );
			break;

		case LS_LOGIN_COMPLETE:
			ASSERT( false );
			pSession->CloseSession();
			break;

		case LS_RELOGIN_CONNECTING:
		case LS_RELOGIN_SEND_DATA:
		case LS_RELOGIN_RECV_DATA:
			OnInternalReloginResult( RC_SOCKET_CLOSED, nSocketError );
			break;
		}
	}
	else if( m_pMainServerNpSession == pSession )
	{
		switch( USER()->m_eLoginStatus )
		{
		case LS_LOGIN_READY:
			ASSERT( false );
			pSession->CloseSession();
			break;

		case LS_LOGIN_CONNECTING:
		case LS_LOGIN_SEND_DATA:
		case LS_LOGIN_RECV_DATA:
			ASSERT( false );
			pSession->CloseSession();
			break;

		case LS_LOGIN_COMPLETE:
			m_pEvents->NpMsgr_OnDisconnected( this, 1);
			
			pSession->CloseSession();
			OnInternalMainSessionClosed(nSocketError);			
			break;

		case LS_RELOGIN_CONNECTING:
		case LS_RELOGIN_SEND_DATA:
		case LS_RELOGIN_RECV_DATA:
			ASSERT( false );
			pSession->CloseSession();
			break;
		}
	}
	else
	{
		ASSERT( false );
	}
}


void CNpMsgrMain::StartCheckNetworkProcess()
{
	TIMER_START( IDT_CHECK_NETWORK, NETWORK_CHECK_TIME, NULL );
}


void CNpMsgrMain::StopCheckNetworkProcess()
{
	TIMER_KILL( IDT_CHECK_NETWORK );
}

void CNpMsgrMain::StartReloginProcess(bool directConnectMode)
{
	//GT_FLOG(L"CNpMsgrMain::StartReloginProcess() -> StartReloginProcess()");

	if(m_nReloginTryCount > 0 && directConnectMode != true)
	{
		//GT_FLOG(L"CNpMsgrMain::StartReloginProcess() -> m_nReloginTryCount > 0 : Already doing...");
		return; //Already doing...
	}

	Disconnect();

	m_nReloginTryCount = 1;

	if( directConnectMode )
	{
		TIMER_START( IDT_RELOGIN_START, 10, NULL );	// Short Interval for fast connect.
//		NTALK_LOG->Debug(L"[LOGIN] IDT_RELOGIN_START [10]");
	}
	else
	{
		UINT uElapse = 3000 + (rand() % (60 * 1000));
		TIMER_START( IDT_RELOGIN_START, uElapse, NULL );
//		NTALK_LOG->Debug(L"[LOGIN] IDT_RELOGIN_START [%d]", uElapse);
	}

	USER()->m_eLoginStatus = LS_RELOGIN_WAIT_START;	
    
    m_bReloginProcess = true;
}


void CNpMsgrMain::StopReloginProcess()
{
    m_bReloginProcess = false;
    
    TIMER_KILL( IDT_RELOGIN_START );
}


void CALLBACK CNpMsgrMain::OnInternalMainSessionClosed(int nSocketError)
{
	char szMsg[1024];

	SYSTEMTIME st;
	GetLocalTime( &st );
	_snprintf(szMsg, 1024, "[%02d%02d %02d:%02d %02d]    ϴ. socket error = %d\r\n", 
		st.wMonth, st.wDay, st.wHour, st.wMinute, st.wSecond, nSocketError );
	m_strLastErrorMsg += szMsg;
	

	// ٷ closed event ó
	//NpSession_AsyncDelete( m_pMainServerNpSession );
	//m_pMainServerNpSession = NULL;	
	//MAINSERVER_SD()->EndMainServerSession();

	//USER()->m_eLoginStatus = LS_KILL_NOTIFY;

	//m_pEvents->NpMsgr_OnKilled( this, 1 );


	// internal relogin
	NpSession_AsyncDelete( m_pMainServerNpSession );
	m_pMainServerNpSession = NULL;
	MAINSERVER_SD()->SetClosedOnlineMode();

	StartReloginProcess(false);
}


void CALLBACK CNpMsgrMain::OnNpSession_RecvPacket(INpSession *pSession, INpPacket * pPacket)
{
//	F_TRACE("CNpMsgrMain::OnNpSession_RecvPacket");

	// Any packet recved.. so current network status is normal! 
	// m_bRecvPongFlag is true..
	m_bRecvPongFlag = true;

	if( !pPacket->IsM6V2() )
	{		
		BYTE byte = *pPacket->RawHeader();
		if(byte == 0x81)	// recv PING
		{
			pSession->SendNpPacket_SessionControlPacket(0x82, false);
		}
		else if(byte == 0x82)
		{
//			m_bRecvPongFlag = true;
			m_nLastRecvPongMSec = GetTickCount();
		}
		return;
	}

	if( pSession == m_pLoginNpSession )
	{
		switch( USER()->m_eLoginStatus )
		{
		case LS_LOGIN_READY:
		case LS_LOGIN_CONNECTING:
			ASSERT( false );
			pSession->CloseSession();
			break;

		case LS_LOGIN_SEND_DATA:
			USER()->m_eLoginStatus = LS_LOGIN_RECV_DATA;
			//  Ʒ ̾..
		case LS_LOGIN_RECV_DATA:
			if( pPacket->M6V2()->usSessionId == 1 )
			{
				// contunue data 
				m_vecLoginResponseData.insert( m_vecLoginResponseData.end(), 
					pPacket->BodyData(),
					pPacket->BodyData() + pPacket->BodyLength() );
			}
			else
			{
				MAINSERVER_SD()->StartMainServerSession( m_pNpMsgrUser, m_pLoginNpSession, pPacket->M6V2()->ulTrId + 1 );

				if( m_vecLoginResponseData.empty() )
				{
					ParseLoginResponsePacket( pPacket->BodyData(), pPacket->BodyLength(), false );
				}
				else
				{
					m_vecLoginResponseData.insert( m_vecLoginResponseData.end(), 
						pPacket->BodyData(),
						pPacket->BodyData() + pPacket->BodyLength() );

					ParseLoginResponsePacket( &m_vecLoginResponseData[0], m_vecLoginResponseData.size(), false );
				}

			}
			break;

		case LS_LOGIN_COMPLETE:
			ASSERT( false );
			pSession->CloseSession();
			break;

		case LS_RELOGIN_CONNECTING:
			// ̷ Ȳ ߻ϸ ȵ.
			ASSERT( false );
			pSession->CloseSession();
			break;

		case LS_RELOGIN_SEND_DATA:
			USER()->m_eLoginStatus = LS_RELOGIN_RECV_DATA;
			ParseLoginResponsePacket( pPacket->BodyData(), pPacket->BodyLength(), true );
			break;
		case LS_RELOGIN_RECV_DATA:
			// ̷ Ȳ ߻ϸ ȵ.
			ASSERT( false );
			pSession->CloseSession();
			break;

		default:
			ASSERT( false );
			pSession->CloseSession();
		}
	}
	else if( m_pMainServerNpSession == pSession )
	{
		switch( USER()->m_eLoginStatus )
		{
		case LS_LOGIN_READY:
		case LS_LOGIN_CONNECTING:
		case LS_LOGIN_SEND_DATA:
		case LS_LOGIN_RECV_DATA:
			ASSERT( false );
			pSession->CloseSession();
			break;

		case LS_LOGIN_COMPLETE:
			if( pPacket->M6V2()->usCMMD == MSNG_CMD_KILL ) //(0x0103)
			{
				// on receive KILL
				pSession->CloseSession();

				NpSession_AsyncDelete( m_pMainServerNpSession );
				m_pMainServerNpSession = NULL;

				m_pNpSD_MainServer->EndMainServerSession();

				USER()->m_eLoginStatus = LS_KILL_NOTIFY;
				m_pEvents->NpMsgr_OnKilled( this, 0 );
				USER()->m_eOnlineStatus = ST_OFFLINE;
				USER()->m_eLoginStatus = LS_LOGIN_READY;

//#if defined(WIN32)				
//				MessageBox(NULL,_T("ٸ PC ID α Ͽϴ.\r\n PC α׾ƿ ˴ϴ."), _T("Ntalk"), MB_OK);
//#endif				
				//Logout();		-> user  NpMsgr_OnKilled ̺Ʈ Logout()  ȣ ؾ..
				// ⼭  ʴ  UI ʿ  ټ ֱ ̴.
			}
			else
				MAINSERVER_SD()->OnReceivePacket( pPacket );
			break;

		default:
			ASSERT( false );
			pSession->CloseSession();
			break;
		}
	}
	else
	{
		ASSERT( false );
	}
}


void CNpMsgrMain::ParseLoginResponsePacket(LPCSTR szData, int nDataLen, bool bRelogin)
{
//	F_TRACE("CNpMsgrMain::ParseLoginResponsePacket(szData=%p, nDataLen=%d) ########################################", szData, nDataLen);

	USHORT usServerResCode;

	try 
	{
#if defined(WIN32)
		USES_CONVERSION;
#endif
		CGBStream stream;
		stream.UseExternalBuffer(szData, nDataLen);

		usServerResCode = stream.Read_U2();	

		// ּ ͸  ϶, ̻ ۾ ʰ Ư ڵ带  .
		// 2   ġ  ̻ 2Ʈ ڵ  ְ Ǿ.
		if(nDataLen == 2)
		{
            if(!bRelogin)
                OnInternalLoginResult( RC_INVALID_DATA, usServerResCode);
            else
                OnInternalReloginResult( RC_INVALID_DATA, usServerResCode );

			return;
		}

		char szErrorString[1024] = {0, };
		WSTRING wzErrorString;

		//   UTF8 ڿ ޵ Ǿ. 
		//  ϴ UTF8 ް CP949 ٽ !
		stream.Read_U8S_to_Unicode_stl(wzErrorString);

		_snprintf(szErrorString, sizeof(szErrorString)-1, "%s", W2A(wzErrorString.c_str()));

//		DLOG( "login server - response code = 0x%04x", usServerResCode );

		switch( usServerResCode )
		{
		case MSNG_OK:
		case MSNG_OK_UPDATECACHE:
		case MSNG_OK_FIRSTLOGIN:
		case MSNG_OK_SAVE_MESSAGE:
			if( !bRelogin )
				ParseLoginSuccessResponse( usServerResCode, stream );
			else
				ParseReloginSuccessResponse( usServerResCode, stream );
			break;
		case MSNG_ERR_RECONNECTNEWJS:
			{
				UINT ulNewServerIP = stream.Read_U4();
				USHORT usNewServerPort = stream.Read_U2();

				char szNewServerIP[32];
				strcpy( szNewServerIP, inet_ntoa( *(in_addr*)&ulNewServerIP ) );

				// internal login retry

				m_pLoginNpSession->CloseSession();

				if( !bRelogin )
					USER()->m_eLoginStatus = LS_LOGIN_CONNECTING;
				else
					USER()->m_eLoginStatus = LS_RELOGIN_CONNECTING;

				m_pLoginNpSession->ConnectSession( szNewServerIP, usNewServerPort );
			}
			break;

		case MSNG_ERR_UPDATEAUTHKEY:
			{
				// update authentication encrypt key cache
				ULONG ulPwdCryptKeyBaseSerial = stream.Read_U4();

				USHORT usPwdCryptKeyBaseLength = stream.Read_U2();
				char * pszPwdCryptKeyBase = (char*)malloc( usPwdCryptKeyBaseLength + 1 );
				if (NULL == pszPwdCryptKeyBase)
				{
					break;
				}

				stream.Read( pszPwdCryptKeyBase, usPwdCryptKeyBaseLength );
				pszPwdCryptKeyBase[usPwdCryptKeyBaseLength] = 0;

				char szPwdCryptKeyBaseSerial[32];
				sprintf(szPwdCryptKeyBaseSerial, "%d", (int)ulPwdCryptKeyBaseSerial);
#if defined(WIN32)	
				WriteRegistryData("PWD Crypt Key Base Serial", (LPBYTE)szPwdCryptKeyBaseSerial, strlen( szPwdCryptKeyBaseSerial ), false, "Info");
				WriteRegistryData("PWD Crypt Key Base", (LPBYTE)pszPwdCryptKeyBase, usPwdCryptKeyBaseLength, false, "Info");
#endif

				free( pszPwdCryptKeyBase );
				if( !bRelogin )
					Send_LOGIN();			
				else
					Send_RELOGIN();
			
			}
			break;

		case MSNG_ERR_FAILAUTH:
			{				
				//m_strLastErrorMsg += string("Server Error Code Description : ") + GetNpIMErrorMsg( usServerResCode );
				m_strLastErrorMsg += szErrorString;		
                
                if( !bRelogin )
                    OnInternalLoginResult( RC_RECV_ERROR_RESULT, usServerResCode );
                else
                    OnInternalReloginResult( RC_RECV_ERROR_RESULT, usServerResCode );
			}
			break;

		case MSNG_ERR_SERVER_DOWN:
			{		
                if( !bRelogin )
                    OnInternalLoginResult( RC_RECV_ERROR_RESULT, usServerResCode );
                else
                    OnInternalReloginResult( RC_RECV_ERROR_RESULT, usServerResCode );
			}
			break;

		case MSNG_ERR_LAUNCHING_CHECK_FAIL:
			{
				USER()->m_nLaunchingResultBits = stream.Read_U2();
				stream.Read_S_stl( USER()->m_strLaunchingResultURL );
				
				if(USER()->m_nLaunchingResultBits != 0)
				{
					m_strLastErrorMsg += "Launching Check Failed";
                    if( !bRelogin )
                        OnInternalLoginResult( RC_RECV_ERROR_RESULT, usServerResCode );
                    else
                        OnInternalReloginResult( RC_RECV_ERROR_RESULT, usServerResCode );
					return;
				}
			}
			
		default:
			if( !bRelogin )
			{
				m_strLastErrorMsg += string("Server Error Code Description : ") + GetNpIMErrorMsg( usServerResCode );				
				OnInternalLoginResult( RC_RECV_ERROR_RESULT, usServerResCode );
			}
			else
			{								
				OnInternalReloginResult( RC_RECV_ERROR_RESULT, usServerResCode );
			}
			break;
		}
	}
	catch( CGBStreamException e )
	{
		if(nDataLen < 2)
		{
            if( !bRelogin )
                OnInternalLoginResult( RC_RECV_ERROR_RESULT, MSNG_ERR_CRACKED_DATA );
            else
				OnInternalReloginResult( RC_RECV_ERROR_RESULT, MSNG_ERR_CRACKED_DATA );
		}
		else
		{
			if( !bRelogin )
				OnInternalLoginResult( RC_GBSTREAM_PARSE_ERROR, e.Error() );
			else
				OnInternalReloginResult( RC_GBSTREAM_PARSE_ERROR, e.Error() );
		}
	}
}

void CNpMsgrMain::ParseLoginSuccessResponse( USHORT usServerResCode, CGBStream & stream )
{
	try
	{
		{
			USHORT usRecordLength = stream.Read_U2();

			USHORT usReadRecordLength = 0;

			usReadRecordLength += stream.Read_U8S_to_Unicode_stl( USER()->m_wstrName );
						
			usReadRecordLength += stream.Read_U8S_to_Unicode_stl( USER()->m_wstrNick );		// ִ°.

			usReadRecordLength += stream.Read_U8S_to_Unicode_stl( USER()->m_wstrId );

			USES_UTF8_CONVERSION
//			DLOG("nick=%s", W2U8(USER()->Nick()) );
//			DLOG("name=%s", W2U8(USER()->Name()) );
//			DLOG("userid=%s", W2U8(USER()->Id()) );

//			NTALK_LOG->Debug(L"[LOGIN] LOGIN_SUCC: [%s]", USER()->Id());

			if( usRecordLength < usReadRecordLength )
			{
				m_strLastErrorMsg += "USER Profile Ϳ record ̺  ̰ .";
				OnInternalLoginResult( RC_INVALID_PROTOCOL, 0 );
				return;
			}
			else if( usRecordLength > usReadRecordLength )
				stream.Skip( usRecordLength - usReadRecordLength );
		}

		{
			USHORT usArrayCount = stream.Read_U2();
//			DLOG("buddy count=%s", usArrayCount );
			if( !BUDDY()->Init_BuddyList( 0/*ulBuddyListCacheSerial*/, usArrayCount, stream ) )
			{
				m_strLastErrorMsg += "Buddy List parse error";
				OnInternalLoginResult( RC_GBSTREAM_PARSE_ERROR, 0 );
				return;
			}
		}
		// web   cookie
		std::string strTicket;
		stream.Read_S_stl( strTicket );
		
		m_pNpMsgrUser->SetAuthCookie(strTicket.c_str());

		USES_UTF8_CONVERSION;
		
//		DLOG("NPMsgrMain Login Success : date size = %ld\n", stream.GetDataLen());
//		DLOG("LS_LOGIN_RECV_DATA , parse complete ########################################");

		//m_nLRSCookie = ulLRSCookie;

		// login  complete Ǿ   online  ƴϴ.
		//     ¸ online  ϰ ģ  notify   ۾   ̴.
		//  ۾ MSNG_CMD_INITONLINE   ̷.
		// UI layer  α  ٸ ߰ ۾  ģ  ¶ ° Ǿ
		// MSNG_CMD_INITONLINE   ȴ. -  CNpMsgrMain::InitOnlineStatus()  ȣ Ͻÿ.
		OnInternalLoginResult( RC_OK, usServerResCode );
	}
	catch( CGBStreamException e )
	{
		OnInternalLoginResult( RC_GBSTREAM_PARSE_ERROR, e.Error() );
	}
}

void CNpMsgrMain::ParseReloginSuccessResponse( USHORT usServerResCode, CGBStream & stream )
{
	ASSERT( USER()->m_eLoginStatus == LS_RELOGIN_RECV_DATA );
	try
	{
		USHORT usRecordLength = stream.Read_U2();

		USHORT usReadRecordLength = 0;
		usReadRecordLength += stream.Read_U8S_to_Unicode_stl( USER()->m_wstrName );
		usReadRecordLength += stream.Read_U8S_to_Unicode_stl( USER()->m_wstrNick );
		usReadRecordLength += stream.Read_U8S_to_Unicode_stl( USER()->m_wstrId );

		USHORT usArrayCount = stream.Read_U2();
//		DLOG("buddy count=%s", usArrayCount );
		if( !BUDDY()->Init_BuddyList( 0, usArrayCount, stream ) )
		{
			m_strLastErrorMsg += "Buddy List parse error";
			OnInternalReloginResult( RC_GBSTREAM_PARSE_ERROR, 0 );
			return;
		}

		// web   cookie
		std::string strTicket;
		stream.Read_S_stl( strTicket );
		
		m_pNpMsgrUser->SetAuthCookie(strTicket.c_str());

	}
	catch( CGBStreamException e )
	{
		m_strLastErrorMsg += "relogin  , web   cookie  дٰ GBStream parse error!";
		OnInternalReloginResult( RC_GBSTREAM_PARSE_ERROR, e.Error() );
		return;
	}

	if( m_pMainServerNpSession )
		m_pMainServerNpSession->Delete();
	m_pMainServerNpSession = m_pLoginNpSession;
	m_pLoginNpSession = NULL;

	MAINSERVER_SD()->RestartMainServerSession( m_pMainServerNpSession, 3 );

	// send to server ADDNOTIFYBUDDY
	//Send_ADDNOTIFYBUDDY();

	// send MSNG_CMD_INITONLINE
	{
		ASSERT( USER()->m_eOnlineStatus != ST_OFFLINE );

		CGBStream stream_send;
		stream_send.CreateStream();
		stream_send.Write_U1( 1 | 2 | 8);
		stream_send.Write_U1( USER()->m_eOnlineStatus );
		stream_send.Write_U8S_from_Unicode( USER()->Nick() );
		
		MAINSERVER_SD()->Send_REQUEST_PACKET( static_cast<USHORT>(MSNG_CMD_INITONLINE), 
			static_cast<USHORT>(stream_send.GetDataLen()),
			stream_send.GetData(), NULL );
	}

	OnInternalReloginResult( RC_OK, usServerResCode );
}


/*
*  Լ MSNG_CMD_ADDNOTIFYBUDDY command   .
* MSNG_CMD_ADDNOTIFYBUDDY  ʿ  buddy list  client  local  cache ϱ ̴.
* client cache  ġ   DB   buddy list   ʿ䰡 .
* ׷. buddy list  PDS(Presence Dispatch Server)  presence notify  ؼ ʿϱ ,
* client  buddy list cache data MSNG_CMD_ADDNOTIFYBUDDY command  ؼ PDS  Ѵ.
*/
void CNpMsgrMain::Send_ADDNOTIFYBUDDY()
{
	CGBStream stream;
	stream.CreateStream(65535);

	IBuddyEnum * pBuddyEnum = BUDDY()->GetAllBuddys( BTYPE_FL | BTYPE_BL );
	if (NULL == pBuddyEnum)
		return;

	stream.Write_U2( pBuddyEnum->Getlength() ); //array count

	for( int i=0 ; i < pBuddyEnum->Getlength() ; i++ )
	{
		IBuddy * pBuddy = pBuddyEnum->Getitem(i);

		UTF8 * pUTF8_id = pBuddy->Id_UTF8();
		if (NULL == pUTF8_id)
			return;

		// record length
		USHORT usRecordLength = 1 + (1+ strlen((LPCSTR)pUTF8_id) );
		stream.Write_U2( usRecordLength );

		stream.Write_U1( pBuddy->BuddyType() );

		stream.Write_U8SS( pUTF8_id );
	}
	pBuddyEnum->Delete();

	MAINSERVER_SD()->Send_REQUEST_PACKET( static_cast<USHORT>(MSNG_CMD_ADDNOTIFYBUDDY), 
		static_cast<USHORT>(stream.GetDataLen()), stream.GetData() );
}


void CALLBACK CNpMsgrMain::OnNpSession_SendReady(INpSession *pSession)
{
}

const char* CNpMsgrMain::GetAppandOSVersionString(char* szBuf, int nBufLen)
{
	string strVersionStr = GetVersionStr();
	strVersionStr += "|";
	strVersionStr += GetOsVersion();

	strcpy_s(szBuf, nBufLen, strVersionStr.c_str());

	return strVersionStr.c_str();
}


/////////////////////////////////////////////////////////////////////////////////
// class CNpSessionDispatcher

void CNpSessionDispatcher::Send_PING(void)
{
	if( !m_bClosedOnlineMode )
	{
		if( !m_pMainServerSession )
			return;

		m_pMainServerSession->SendNpPacket_SessionControlPacket(0x81, false);
	}
	else
	{
		// Don't need to send ping..
	}
}

bool CNpSessionDispatcher::PING()
{
	if( m_bClosedOnlineMode )
		return false;
	
	if( m_pMainServerSession == NULL )
		return false;

	return m_pMainServerSession->SendNpPacket_SessionControlPacket(0x81, false);
}

void CNpSessionDispatcher::Register_INCOMING_REQUEST_PACKET_Handler(USHORT usCMMD, _INpSessionDispatcher_PacketHandler * pHandler)
{
	ASSERT( m_mapReqCMMDHandler.find( usCMMD ) == m_mapReqCMMDHandler.end() );

	m_mapReqCMMDHandler.insert( 
		map<USHORT, _INpSessionDispatcher_PacketHandler*>::value_type( usCMMD, pHandler ) );
}

void CNpSessionDispatcher::Deregister_INCOMING_REQUEST_PACKET_Handler(USHORT usCMMD)
{
	m_mapReqCMMDHandler.erase( usCMMD );
}

int CNpSessionDispatcher::Send_REQUEST_PACKET(USHORT usCMMD, 
											  USHORT usBodyLength, 
											  LPCSTR szBodyData, 
											  _INpSessionDispatcher_PacketHandler * pRESPONSE_PACKET_Handler)	// return trid
{
	int nTrId = m_nTrId_Serial++;

	if( m_bClosedOnlineMode )
	{
		// buffering
		m_listClosedOnlineSendBuffer.push_back( 
			new tCLOSED_ONLINE_SEND_BUFFER( usCMMD, nTrId, szBodyData, usBodyLength) );
	}
	else
	{
		if( !m_pMainServerSession )
			return -1;

		m_pMainServerSession->SendNpPacket_M6V2( usCMMD, nTrId, usBodyLength, szBodyData );
	}

	if( pRESPONSE_PACKET_Handler )
		m_mapResTRIDHandler.insert( map<int, _INpSessionDispatcher_PacketHandler*>::value_type( nTrId, pRESPONSE_PACKET_Handler ) );

	char strTrid[100];
	wsprintf(strTrid, "Send_REQUEST_PACKET TrId :%ld\r\n", nTrId);
	OutputDebugString(strTrid);

	return nTrId;
}
bool CNpSessionDispatcher::Deregister_RESPONSE_PACKET_Handler(int nTrId)
{
	m_mapResTRIDHandler.erase( nTrId );
	return true;
}

void CNpSessionDispatcher::Send_RESPONSE_PACKET(USHORT usCMMD, int nTrId, USHORT usBodyLength, LPCSTR szBodyData)	// return trid
{
	ASSERT( !IS_REQUEST_CMMD( usCMMD ) );

	if( m_bClosedOnlineMode )
	{
		// buffering
		m_listClosedOnlineSendBuffer.push_back( 
			new tCLOSED_ONLINE_SEND_BUFFER( usCMMD, nTrId, szBodyData, usBodyLength) );
	}
	else
	{
		if( !m_pMainServerSession )
			return;

		m_pMainServerSession->SendNpPacket_M6V2( usCMMD, nTrId, usBodyLength, szBodyData );
	}

}

void CNpSessionDispatcher::Register_CToC_PACKET_Handler(USHORT usCToC_SubCMMD, _INpSessionDispatcher_PacketHandler * pHandler)
{
	//ASSERT( m_mapCToC_SubCMMDHandler.find( usCToC_SubCMMD ) == m_mapCToC_SubCMMDHandler.end() );

	map<USHORT, list<_INpSessionDispatcher_PacketHandler*> >::iterator it = m_mapCToC_SubCMMDHandler.find( usCToC_SubCMMD );
	
	if(it != m_mapCToC_SubCMMDHandler.end())
	{
		it->second.push_back(pHandler);
	}
	else
	{
		list<_INpSessionDispatcher_PacketHandler *> listHandler;
		
		listHandler.push_back(pHandler);
		
		m_mapCToC_SubCMMDHandler.insert(map<USHORT, list<_INpSessionDispatcher_PacketHandler*> >::value_type( usCToC_SubCMMD, listHandler ) );
	}
	
	
}

void CNpSessionDispatcher::Deregister_CToC_PACKET_Handler(USHORT usCToC_SubCMMD, _INpSessionDispatcher_PacketHandler *pHandler)
{
	map<USHORT, list<_INpSessionDispatcher_PacketHandler*> >::iterator it = m_mapCToC_SubCMMDHandler.find( usCToC_SubCMMD );
	
	if(it != m_mapCToC_SubCMMDHandler.end())
	{
		it->second.remove(pHandler);
		
		if(it->second.size() <= 0)
		{
			m_mapCToC_SubCMMDHandler.erase( usCToC_SubCMMD );
		}
	}
}

//int  CNpSessionDispatcher::Send_CToC_PACKET(LPCWSTR wszToId, USHORT usBodyLength, LPCSTR szBodyData, _INpSessionDispatcher_PacketHandler * pRESPONSE_PACKET_Handler)	// return trid
//{
//	int nTrId = m_nTrId_Serial++;
//
//	if( m_bClosedOnlineMode )
//	{
//		// buffering
//		tCLOSED_ONLINE_SEND_BUFFER * pBufferData = 
//			new tCLOSED_ONLINE_SEND_BUFFER( 0, nTrId, szBodyData, usBodyLength);
//
//		pBufferData->m_wstrToId = wszToId;
//
//		m_listClosedOnlineSendBuffer.push_back( pBufferData );
//	}
//	else
//	{
//		if( !m_pMainServerSession )
//			return -1;
//
//		m_pMainServerSession->SendNpPacket_M6V2( MSNG_CMD_SENDTOCLIENT,
//			nTrId, 
//			usBodyLength, 
//			szBodyData, 
//			(_t_UTF8)wszToId, 
//			(_t_UTF8)m_pNpMsgrUser->Id() );
//	}
//
//
//	if( pRESPONSE_PACKET_Handler )
//		m_mapResTRIDHandler.insert( map<int, _INpSessionDispatcher_PacketHandler*>::value_type( nTrId, pRESPONSE_PACKET_Handler ) );
//
//	return nTrId;
//}


void CALLBACK CNpSessionDispatcher::OnReceivePacket(const INpPacket * pPacket)
{
	USHORT usCMD = pPacket->M6V2()->usCMMD;
	
	//if( usCMD == MSNG_CMD_SENDTOCLIENT )
	//{
	//	USHORT usCToC_SubCMMD;

	//	if( pPacket->BodyLength() < 2 )
	//		return;

	//	memcpy( &usCToC_SubCMMD, pPacket->BodyData(), 2 );

	//	map<USHORT, list<_INpSessionDispatcher_PacketHandler*> >::iterator it = m_mapCToC_SubCMMDHandler.find( usCToC_SubCMMD );
	//	if( it != m_mapCToC_SubCMMDHandler.end() )
	//	{
	//		list<_INpSessionDispatcher_PacketHandler*>::iterator it2 = it->second.begin();
	//		for(; it2 != it->second.end(); it2++)
	//		{
	//			_INpSessionDispatcher_PacketHandler * pHandler = *it2;
	//			pHandler->INpSessionDispatcher_OnReceive_CToC( this, usCToC_SubCMMD, pPacket );
	//		}
	//	}
	//	else
	//	{
	//		IK_TRACE("NpMsgr.dll: receive unhandled CToC, sub cmmd=0x%04x\n", usCToC_SubCMMD );
	//	}

	//	return;
	//}

	if( IS_REQUEST_CMMD( usCMD ) )
	{
		char strDebug[1024];
		wsprintf(strDebug, "usCMMD_1 : %ld(%04x)", usCMD, usCMD);
		OutputDebugString(strDebug);

		map<USHORT, _INpSessionDispatcher_PacketHandler*>::iterator it = m_mapReqCMMDHandler.find( usCMD );
		if( it != m_mapReqCMMDHandler.end() )
		{
			_INpSessionDispatcher_PacketHandler * pHandler = it->second;
			pHandler->INpSessionDispatcher_OnReceive_REQUEST_PACKET( this, pPacket );
		}
		else
		{
//			IK_TRACE("NpMsgr.dll: receive unhandled reqCMMD 0x%04x\n", usCMD );
		}
	}
	else
	{
		char strTrid[100];
		wsprintf(strTrid, "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~OnReceivePacket TrId :%ld\r\n", pPacket->M6V2()->ulTrId);
		OutputDebugString(strTrid);

		map<int, _INpSessionDispatcher_PacketHandler*>::iterator it = m_mapResTRIDHandler.find( pPacket->M6V2()->ulTrId );
		if( it != m_mapResTRIDHandler.end() )
		{
			_INpSessionDispatcher_PacketHandler * pHandler = it->second;
			pHandler->INpSessionDispatcher_OnReceive_RESPONSE_PACKET( this, NULL, pPacket );
		}
		else
		{
			char strDebug[1024];
			wsprintf(strDebug, "usCMMD_2 : %ld(%04x)", usCMD, usCMD);
			OutputDebugString(strDebug);

//			IK_TRACE("NpMsgr.dll: receive unhandled reseponse 0x%04x\n", usCMD );
		}
	}
}

void CNpSessionDispatcher::Send_ClosedOnlineSendBuffer()
{
	while( !m_listClosedOnlineSendBuffer.empty() )
	{
		tCLOSED_ONLINE_SEND_BUFFER* pBufferData = m_listClosedOnlineSendBuffer.front();
		m_listClosedOnlineSendBuffer.pop_front();

		if( pBufferData->m_wstrToId.empty() )
		{
			//   Ŷ
			m_pMainServerSession->SendNpPacket_M6V2( pBufferData->m_usCMMD,
				pBufferData->m_nTrId,
				pBufferData->m_nDataLen,
				pBufferData->m_pData );
		}
		//else
		//{
		//	// CtoC Ŷ..

		//	m_pMainServerSession->SendNpPacket_M6V2( MSNG_CMD_SENDTOCLIENT,
		//		pBufferData->m_nTrId,
		//		pBufferData->m_nDataLen,
		//		pBufferData->m_pData, 
		//		(_t_UTF8)pBufferData->m_wstrToId.c_str(), 
		//		(_t_UTF8)m_pNpMsgrUser->Id() );
		//}
	}
}

void CNpSessionDispatcher::Clear_ClosedOnlineSendBuffer()
{
	list<tCLOSED_ONLINE_SEND_BUFFER*>::iterator it = m_listClosedOnlineSendBuffer.begin();
	for( ; it != m_listClosedOnlineSendBuffer.end() ; it++ )
	{
		tCLOSED_ONLINE_SEND_BUFFER* p = *it;

		delete p;
	}

	m_listClosedOnlineSendBuffer.clear();
}
